import logging
from typing import Type, Tuple, List

from peek_storage_service._private.plugin.ServerFrontendLoadersMixin import (
    ServerFrontendLoadersMixin,
)
from twisted.internet.defer import inlineCallbacks

from peek_platform.plugin.PluginLoaderABC import PluginLoaderABC
from peek_plugin_base.PluginCommonEntryHookABC import PluginCommonEntryHookABC
from peek_plugin_base.server.PluginLogicEntryHookABC import PluginLogicEntryHookABC
from peek_plugin_base.server.PluginServerStorageEntryHookABC import (
    PluginServerStorageEntryHookABC,
)
from peek_plugin_base.server.PluginServerWorkerEntryHookABC import (
    PluginServerWorkerEntryHookABC,
)
from peek_storage_service._private.plugin.PeekStoragePlatformHook import (
    PeekStoragePlatformHook,
)

logger = logging.getLogger(__name__)


class StoragePluginLoader(PluginLoaderABC, ServerFrontendLoadersMixin):
    _instance = None

    def __new__(cls, *args, **kwargs):
        assert (
            cls._instance is None
        ), "StoragePluginLoader is a singleton, don't construct it"
        cls._instance = PluginLoaderABC.__new__(cls, *args, **kwargs)
        return cls._instance

    def __init__(self, *args, **kwargs):
        PluginLoaderABC.__init__(self, *args, **kwargs)

    @property
    def _entryHookFuncName(self) -> str:
        return "peekStorageEntryHook"

    @property
    def _entryHookClassType(self):
        return PluginLogicEntryHookABC

    @property
    def _platformServiceNames(self) -> List[str]:
        return ["logic", "storage"]

    @inlineCallbacks
    def loadOptionalPlugins(self):
        yield PluginLoaderABC.loadOptionalPlugins(self)

    def unloadPlugin(self, pluginName: str):
        PluginLoaderABC.unloadPlugin(self, pluginName)

        self._unloadPluginFromAdminSite(pluginName)

    @inlineCallbacks
    def _loadPluginThrows(
        self,
        pluginName: str,
        EntryHookClass: Type[PluginCommonEntryHookABC],
        pluginRootDir: str,
        requiresService: Tuple[str, ...],
    ) -> PluginCommonEntryHookABC:
        # Everyone gets their own instance of the plugin API
        platformApi = PeekStoragePlatformHook(pluginName)

        pluginMain = EntryHookClass(
            pluginName=pluginName, pluginRootDir=pluginRootDir, platform=platformApi
        )

        # Load the plugin
        yield pluginMain.load()

        if (
            not isinstance(pluginMain, PluginServerWorkerEntryHookABC)
            and "worker" in requiresService
        ):
            raise Exception(
                "Plugin %s requires 'worker' service."
                " It must now inherit PluginServerWorkerEntryHookABC"
                " in its PluginLogicEntryHook implementation"
            )

        if isinstance(pluginMain, PluginServerStorageEntryHookABC):

            metadata = pluginMain.dbMetadata
            schemaName = pluginName.replace("peek_plugin_", "pl_").replace(
                "peek_core_", "core_"
            )
            if metadata.schema != schemaName:
                raise Exception(
                    "Peek plugin %s db schema name is %s, should be %s"
                    % (pluginName, metadata.schema, schemaName)
                )

            # Create/Migrate the database schema
            pluginMain._migrateStorageSchema(pluginMain.dbMetadata)

        # Check the implementation
        elif "storage" in requiresService:
            raise Exception(
                "Plugin %s requires 'storage' service."
                " It must now inherit PluginServerStorageEntryHookMixin"
                " in its PluginLogicEntryHook implementation"
            )

        # Add all the resources required to serve the backend site
        # And all the plugin custom resources it may create

        self._loadedPlugins[pluginName] = pluginMain



import sys
from typing import Callable, Any, Optional

import json
from peek_plugin_base.storage.DbConnection import DbSessionCreator
from sqlalchemy import func
from vortex.Tuple import addTupleType, TupleField, Tuple

__sysPathsJson = json.dumps(sys.path)


@addTupleType
class _RunPyInPgArgTuple(Tuple):
    __tupleType__ = "peek_storage_service._RunPyInPgArgTuple"
    args = TupleField()
    kwargs = TupleField()


@addTupleType
class _RunPyInPgResultTuple(Tuple):
    __tupleType__ = "peek_storage_service._RunPyInPgResultTuple"
    result = TupleField()


def runPyInPgBlocking(
    dbSessionCreator: DbSessionCreator,
    classMethodToRun: Callable,
    classMethodToImportTuples: Optional[Callable],
    *args,
    **kwargs
) -> Any:
    # noinspection PyProtectedMember
    argTupleJson = _RunPyInPgArgTuple(args=args, kwargs=kwargs)._toJson()

    loaderModuleClassMethodToRunStr = ".".join(
        [
            classMethodToRun.__self__.__module__,
            classMethodToRun.__self__.__name__,
            classMethodToRun.__name__,
        ]
    )

    loaderModuleClassMethodToImportStr = (
        ".".join(
            [
                classMethodToImportTuples.__self__.__module__,
                classMethodToImportTuples.__self__.__name__,
                classMethodToImportTuples.__name__,
            ]
        )
        if classMethodToImportTuples
        else "None"
    )

    session = dbSessionCreator()
    try:
        sqlFunc = func.peek_storage_service.run_generic_python(
            argTupleJson,
            loaderModuleClassMethodToRunStr,
            loaderModuleClassMethodToImportStr,
            __sysPathsJson,
        )

        resultJsonStr: str = next(session.execute(sqlFunc))[0]
        session.commit()

        # noinspection PyProtectedMember
        resultTuple = _RunPyInPgResultTuple()._fromJson(resultJsonStr)

        return resultTuple.result

    finally:
        session.close()



"""up1_run_generic_py

Revision ID: 51e48a6594d5
Revises: 72518909970f
Create Date: 2020-06-02 22:27:48.650874

"""

# revision identifiers, used by Alembic.
revision = "51e48a6594d5"
down_revision = "72518909970f"
branch_labels = None
depends_on = None

from alembic import op


def upgrade():
    sql = """

DROP FUNCTION IF EXISTS
              peek_storage_service.run_generic_python(character varying,
                                              character varying,
                                              character varying,
                                              character varying);

CREATE OR REPLACE FUNCTION peek_storage_service.run_generic_python(
    args_tuple_json_str character varying,
    class_method_to_run_str_ character varying,
    class_method_to_import_tuples_ character varying,
    python_path character varying)
    RETURNS character varying
    LANGUAGE 'plpython3u'

    COST 100
    VOLATILE

AS $BODY$

import json
from base64 import b64decode
argsTupleJsonStr = args_tuple_json_str
classMethodToRunStr = class_method_to_run_str_
classMethodToImportTuplesStr = class_method_to_import_tuples_
pythonPath = json.loads(python_path)

# ---------------
# Setup to use the virtual environment
import sys

sys.path.extend(pythonPath)

from importlib.util import find_spec, module_from_spec

# ---------------
# Dynamically load the import tuple method

if classMethodToImportTuplesStr and classMethodToImportTuplesStr != 'None':
    modName, className, methodName = classMethodToImportTuplesStr.rsplit('.',2)

    if modName in sys.modules:
        package_ = sys.modules[modName]

    else:
        modSpec = find_spec(modName)
        if not modSpec:
             raise Exception("Failed to find package %s,"
                             " is the python package installed?" % modName)
        package_ = modSpec.loader.load_module()

    Class_ = getattr(package_, className)
    importTupleMethod = getattr(Class_, methodName)
    importTupleMethod()

# ---------------
# Dynamically load the tuple create method

modName, className, methodName = classMethodToRunStr.rsplit('.',2)

if modName in sys.modules:
    package_ = sys.modules[modName]

else:
    modSpec = find_spec(modName)
    if not modSpec:
         raise Exception("Failed to find package %s,"
                         " is the python package installed?" % modName)
    package_ = modSpec.loader.load_module()

Class_ = getattr(package_, className)
classMethodToRun = getattr(Class_, methodName)

# ---------------
# Load the arguments

from peek_storage_service.plpython.RunPyInPg import _RunPyInPgResultTuple, _RunPyInPgArgTuple

argsTuple = _RunPyInPgArgTuple()._fromJson(argsTupleJsonStr)

# ---------------
# Run the method

result = classMethodToRun(plpy, *argsTuple.args, **argsTuple.kwargs)

# ---------------
# Return the result

return _RunPyInPgResultTuple(result=result)._toJson()

$BODY$;

ALTER FUNCTION peek_storage_service.run_generic_python(character varying,
                                               character varying,
                                               character varying,
                                               character varying)
    OWNER TO peek;


          """
    op.execute(sql)


def downgrade():
    pass
